@@ -1,6 +1,6 @@
 /*****************************************************************************
 Copyright (C) 2013, 2015, Google Inc. All Rights Reserved.
-Copyright (c) 2014, 2020, MariaDB Corporation.
+Copyright (c) 2014, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -298,9 +298,14 @@ fil_space_read_crypt_data(const page_size_t& page_size, const byte* page)
 	      type == CRYPT_SCHEME_1)
 	    || iv_length != sizeof crypt_data->iv) {
 		ib::error() << "Found non sensible crypt scheme: "
-			    << type << "," << iv_length
-			    << " for space: "
-			    << page_get_space_id(page);
+			    << type << "," << iv_length << " for space: "
+			    << page_get_space_id(page) << " offset: "
+			    << offset << " bytes: ["
+			    << page[offset + 2 + MAGIC_SZ]
+			    << page[offset + 3 + MAGIC_SZ]
+			    << page[offset + 4 + MAGIC_SZ]
+			    << page[offset + 5 + MAGIC_SZ]
+			    << "].";
 		return NULL;
 	}
 
@@ -451,7 +456,6 @@ byte*
 fil_parse_write_crypt_data(
 	byte*			ptr,
 	const byte*		end_ptr,
-	const buf_block_t*	block,
 	dberr_t*		err)
 {
 	/* check that redo log entry is complete */
@@ -498,12 +502,12 @@ fil_parse_write_crypt_data(
 		return NULL;
 	}
 
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 
 	fil_space_t* space = fil_space_get_by_id(space_id);
 
 	if (!space) {
-		mutex_exit(&fil_system->mutex);
+		mutex_exit(&fil_system.mutex);
 		return ptr + len;
 	}
 
@@ -524,7 +528,7 @@ fil_parse_write_crypt_data(
 		space->crypt_data = crypt_data;
 	}
 
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 
 	if (crypt_data->should_encrypt() && !crypt_data->is_key_found()) {
 		*err = DB_DECRYPTION_FAILED;
@@ -585,7 +589,7 @@ fil_encrypt_buf(
 
 	int rc = encryption_scheme_encrypt(src, srclen, dst, &dstlen,
 					   crypt_data, key_version,
-					   space, offset, lsn);
+					   (uint32)space, (uint32)offset, lsn);
 	ut_a(rc == MY_AES_OK);
 	ut_a(dstlen == srclen);
 
@@ -650,10 +654,51 @@ fil_space_encrypt(
 		return (src_frame);
 	}
 
-	ut_ad(space->n_pending_ios > 0);
-	return fil_encrypt_buf(space->crypt_data, space->id, offset, lsn,
-			       src_frame, page_size_t(space->flags),
-			       dst_frame);
+	fil_space_crypt_t* crypt_data = space->crypt_data;
+	const page_size_t	page_size(space->flags);
+	ut_ad(space->pending_io());
+	byte* tmp = fil_encrypt_buf(crypt_data, space->id, offset, lsn,
+				    src_frame, page_size, dst_frame);
+
+#ifdef UNIV_DEBUG
+	if (tmp) {
+		/* Verify that encrypted buffer is not corrupted */
+		dberr_t err = DB_SUCCESS;
+		byte* src = src_frame;
+		bool page_compressed_encrypted = (mach_read_from_2(tmp+FIL_PAGE_TYPE) == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED);
+		byte uncomp_mem[UNIV_PAGE_SIZE_MAX];
+		byte tmp_mem[UNIV_PAGE_SIZE_MAX];
+
+		if (page_compressed_encrypted) {
+			memcpy(uncomp_mem, src, srv_page_size);
+			ulint unzipped1 = fil_page_decompress(
+				tmp_mem, uncomp_mem);
+			ut_ad(unzipped1);
+			if (unzipped1 != srv_page_size) {
+				src = uncomp_mem;
+			}
+		}
+
+		ut_ad(!buf_page_is_corrupted(true, src, page_size, space));
+		ut_ad(fil_space_decrypt(crypt_data, tmp_mem, page_size, tmp,
+					&err));
+		ut_ad(err == DB_SUCCESS);
+
+		/* Need to decompress the page if it was also compressed */
+		if (page_compressed_encrypted) {
+			byte buf[UNIV_PAGE_SIZE_MAX];
+			memcpy(buf, tmp_mem, srv_page_size);
+			ulint unzipped2 = fil_page_decompress(tmp_mem, buf);
+			ut_ad(unzipped2);
+		}
+
+		memcpy(tmp_mem + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION,
+		       src + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 8);
+		ut_ad(!memcmp(src, tmp_mem, page_size.physical()));
+	}
+#endif /* UNIV_DEBUG */
+
+	return tmp;
 }
 
 /** Decrypt a page.
@@ -720,10 +765,8 @@ fil_space_decrypt(
 		}
 
 		ib::fatal() << "Unable to decrypt data-block "
-			    << " src: " << static_cast<const void*>(src)
-			    << "srclen: "
-			    << srclen << " buf: "
-			    << static_cast<const void*>(dst) << "buflen: "
+			    << " src: " << src << "srclen: "
+			    << srclen << " buf: " << dst << "buflen: "
 			    << dstlen << " return-code: " << rc
 			    << " Can't continue!";
 	}
@@ -749,6 +792,7 @@ Decrypt a page.
 @param[in]	space			Tablespace
 @param[in]	tmp_frame		Temporary buffer used for decrypting
 @param[in,out]	src_frame		Page to decrypt
+@param[out]	decrypted		true if page was decrypted
 @return decrypted page, or original not encrypted page if decryption is
 not needed.*/
 UNIV_INTERN
@@ -756,20 +800,23 @@ byte*
 fil_space_decrypt(
 	const fil_space_t* space,
 	byte*		tmp_frame,
-	byte*		src_frame)
+	byte*		src_frame,
+	bool*		decrypted)
 {
 	dberr_t err = DB_SUCCESS;
 	byte* res = NULL;
 	const page_size_t page_size(space->flags);
+	*decrypted = false;
 
 	ut_ad(space->crypt_data != NULL && space->crypt_data->is_encrypted());
-	ut_ad(space->n_pending_ios > 0);
+	ut_ad(space->pending_io());
 
 	bool encrypted = fil_space_decrypt(space->crypt_data, tmp_frame,
 					   page_size, src_frame, &err);
 
 	if (err == DB_SUCCESS) {
 		if (encrypted) {
+			*decrypted = true;
 			/* Copy the decrypted page back to page buffer, not
 			really any other options. */
 			memcpy(src_frame, tmp_frame, page_size.physical());
@@ -912,12 +959,12 @@ fil_crypt_read_crypt_data(fil_space_t* space)
 	mtr.start();
 	if (buf_block_t* block = buf_page_get(page_id_t(space->id, 0),
 					      page_size, RW_S_LATCH, &mtr)) {
-		mutex_enter(&fil_system->mutex);
+		mutex_enter(&fil_system.mutex);
 		if (!space->crypt_data) {
 			space->crypt_data = fil_space_read_crypt_data(
 				page_size, block->frame);
 		}
-		mutex_exit(&fil_system->mutex);
+		mutex_exit(&fil_system.mutex);
 	}
 	mtr.commit();
 }
@@ -970,9 +1017,9 @@ static bool fil_crypt_start_encrypting_space(fil_space_t* space)
 	crypt_data->rotate_state.starting = true;
 	crypt_data->rotate_state.active_threads = 1;
 
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 	space->crypt_data = crypt_data;
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 
 	fil_crypt_start_converting = true;
 	mutex_exit(&fil_crypt_threads_mutex);
@@ -1076,7 +1123,6 @@ struct rotate_thread_t {
 		case SRV_SHUTDOWN_EXIT_THREADS:
 			/* srv_init_abort() must have been invoked */
 		case SRV_SHUTDOWN_CLEANUP:
-		case SRV_SHUTDOWN_INITIATED:
 			return true;
 		case SRV_SHUTDOWN_FLUSH_PHASE:
 		case SRV_SHUTDOWN_LAST_PHASE:
@@ -1107,7 +1153,7 @@ fil_crypt_space_needs_rotation(
 		return false;
 	}
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
@@ -1402,7 +1448,7 @@ static bool fil_crypt_find_space_to_rotate(
 
 	if (state->should_shutdown()) {
 		if (state->space) {
-			fil_space_release(state->space);
+			state->space->release();
 			state->space = NULL;
 		}
 		return false;
@@ -1411,7 +1457,7 @@ static bool fil_crypt_find_space_to_rotate(
 	if (state->first) {
 		state->first = false;
 		if (state->space) {
-			fil_space_release(state->space);
+			state->space->release();
 		}
 		state->space = NULL;
 	}
@@ -1419,13 +1465,10 @@ static bool fil_crypt_find_space_to_rotate(
 	/* If key rotation is enabled (default) we iterate all tablespaces.
 	If key rotation is not enabled we iterate only the tablespaces
 	added to keyrotation list. */
-	if (srv_fil_crypt_rotate_key_age) {
-		state->space = fil_space_next(state->space);
-	} else {
-		state->space = fil_system->keyrotate_next(
-					state->space, *recheck,
-					key_state->key_version);
-	}
+	state->space = srv_fil_crypt_rotate_key_age
+		? fil_space_next(state->space)
+		: fil_system.keyrotate_next(state->space, *recheck,
+					    key_state->key_version);
 
 	while (!state->should_shutdown() && state->space) {
 		/* If there is no crypt data and we have not yet read
@@ -1443,13 +1486,10 @@ static bool fil_crypt_find_space_to_rotate(
 			return true;
 		}
 
-		if (srv_fil_crypt_rotate_key_age) {
-			state->space = fil_space_next(state->space);
-		} else {
-			state->space = fil_system->keyrotate_next(
-					state->space, *recheck,
-					key_state->key_version);
-		}
+		state->space = srv_fil_crypt_rotate_key_age
+			? fil_space_next(state->space)
+			: fil_system.keyrotate_next(state->space, *recheck,
+						    key_state->key_version);
 	}
 
 	/* if we didn't find any space return iops */
@@ -1522,7 +1562,7 @@ fil_crypt_find_page_to_rotate(
 	ulint batch = srv_alloc_time * state->allocated_iops;
 	fil_space_t* space = state->space;
 
-	ut_ad(!space || space->n_pending_ops > 0);
+	ut_ad(!space || space->referenced());
 
 	/* If space is marked to be dropped stop rotation. */
 	if (!space || space->is_stopping()) {
@@ -1580,7 +1620,7 @@ fil_crypt_get_page_throttle_func(
 	fil_space_t* space = state->space;
 	const page_size_t page_size = page_size_t(space->flags);
 	const page_id_t page_id(space->id, offset);
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	/* Before reading from tablespace we need to make sure that
 	the tablespace is not about to be dropped. */
@@ -1662,11 +1702,11 @@ btr_scrub_get_block_and_allocation_status(
 	buf_block_t *block = NULL;
 	fil_space_t* space = state->space;
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	mtr_start(&local_mtr);
 
-	*allocation_status = fseg_page_is_free(space, offset) ?
+	*allocation_status = fseg_page_is_free(space, (uint32_t)offset) ?
 		BTR_SCRUB_PAGE_FREE :
 		BTR_SCRUB_PAGE_ALLOCATED;
 
@@ -1714,7 +1754,7 @@ fil_crypt_rotate_page(
 	ulint sleeptime_ms = 0;
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 	ut_ad(offset > 0);
 
 	/* In fil_crypt_thread where key rotation is done we have
@@ -1730,6 +1770,8 @@ fil_crypt_rotate_page(
 		return;
 	}
 
+	ut_d(const bool was_free = fseg_page_is_free(space, (uint32_t)offset));
+
 	mtr_t mtr;
 	mtr.start();
 	if (buf_block_t* block = fil_crypt_get_page_throttle(state,
@@ -1744,9 +1786,9 @@ fil_crypt_rotate_page(
 		if (space->is_stopping()) {
 			/* The tablespace is closing (in DROP TABLE or
 			TRUNCATE TABLE or similar): avoid further access */
-		} else if (!kv && !*reinterpret_cast<uint16_t*>
-			   (&frame[FIL_PAGE_TYPE])) {
-			/* It looks like this page is not
+		} else if (!*reinterpret_cast<uint32_t*>(FIL_PAGE_OFFSET
+							 + frame)) {
+			/* It looks like this page was never
 			allocated. Because key rotation is accessing
 			pages in a pattern that is unlike the normal
 			B-tree and undo log access pattern, we cannot
@@ -1756,20 +1798,9 @@ fil_crypt_rotate_page(
 			tablespace latch before acquiring block->lock,
 			then the fseg_page_is_free() information
 			could be stale already. */
-
-			/* If the data file was originally created
-			before MariaDB 10.0 or MySQL 5.6, some
-			allocated data pages could carry 0 in
-			FIL_PAGE_TYPE. The FIL_PAGE_TYPE on those
-			pages will be updated in
-			buf_flush_init_for_writing() when the page
-			is modified the next time.
-
-			Also, when the doublewrite buffer pages are
-			allocated on bootstrap in a non-debug build,
-			some dummy pages will be allocated, with 0 in
-			the FIL_PAGE_TYPE. Those pages should be
-			skipped from key rotation forever. */
+			ut_ad(was_free);
+			ut_ad(kv == 0);
+			ut_ad(page_get_space_id(frame) == 0);
 		} else if (fil_crypt_needs_rotation(
 				crypt_data,
 				kv,
@@ -1901,7 +1932,7 @@ fil_crypt_rotate_pages(
 	ulint end = std::min(state->offset + state->batch,
 			     state->space->free_limit);
 
-	ut_ad(state->space->n_pending_ops > 0);
+	ut_ad(state->space->referenced());
 
 	for (; state->offset < end; state->offset++) {
 
@@ -1940,7 +1971,7 @@ fil_crypt_flush_space(
 	fil_space_t* space = state->space;
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	/* flush tablespace pages so that there are no pages left with old key */
 	lsn_t end_lsn = crypt_data->rotate_state.end_lsn;
@@ -1995,18 +2026,13 @@ fil_crypt_flush_space(
 
 /***********************************************************************
 Complete rotating a space
-@param[in,out]		key_state		Key state
 @param[in,out]		state			Rotation state */
-static
-void
-fil_crypt_complete_rotate_space(
-	const key_state_t*	key_state,
-	rotate_thread_t*	state)
+static void fil_crypt_complete_rotate_space(rotate_thread_t* state)
 {
 	fil_space_crypt_t *crypt_data = state->space->crypt_data;
 
 	ut_ad(crypt_data);
-	ut_ad(state->space->n_pending_ops > 0);
+	ut_ad(state->space->referenced());
 
 	/* Space might already be dropped */
 	if (!state->space->is_stopping()) {
@@ -2159,9 +2185,8 @@ DECLARE_THREAD(fil_crypt_thread)(void*)
 				/* If space is marked as stopping, release
 				space and stop rotation. */
 				if (thr.space->is_stopping()) {
-					fil_crypt_complete_rotate_space(
-						&new_state, &thr);
-					fil_space_release(thr.space);
+					fil_crypt_complete_rotate_space(&thr);
+					thr.space->release();
 					thr.space = NULL;
 					break;
 				}
@@ -2172,7 +2197,7 @@ DECLARE_THREAD(fil_crypt_thread)(void*)
 
 			/* complete rotation */
 			if (thr.space) {
-				fil_crypt_complete_rotate_space(&new_state, &thr);
+				fil_crypt_complete_rotate_space(&thr);
 			}
 
 			/* force key state refresh */
@@ -2188,7 +2213,7 @@ DECLARE_THREAD(fil_crypt_thread)(void*)
 
 	/* release current space if shutting down */
 	if (thr.space) {
-		fil_space_release(thr.space);
+		thr.space->release();
 		thr.space = NULL;
 	}
 
@@ -2253,13 +2278,13 @@ fil_crypt_set_thread_cnt(
 if innodb_encryption_rotate_key_age=0. */
 static void fil_crypt_rotation_list_fill()
 {
-	ut_ad(mutex_own(&fil_system->mutex));
+	ut_ad(mutex_own(&fil_system.mutex));
 
-	for (fil_space_t* space = UT_LIST_GET_FIRST(fil_system->space_list);
+	for (fil_space_t* space = UT_LIST_GET_FIRST(fil_system.space_list);
 	     space != NULL;
 	     space = UT_LIST_GET_NEXT(space_list, space)) {
 		if (space->purpose != FIL_TYPE_TABLESPACE
-		    || space->is_in_rotation_list
+		    || space->is_in_rotation_list()
 		    || space->is_stopping()
 		    || UT_LIST_GET_LEN(space->chain) == 0) {
 			continue;
@@ -2268,11 +2293,13 @@ static void fil_crypt_rotation_list_fill()
 		/* Ensure that crypt_data has been initialized. */
 		if (!space->size) {
 			/* Protect the tablespace while we may
-			release fil_system->mutex. */
+			release fil_system.mutex. */
 			space->n_pending_ops++;
-			ut_d(const fil_space_t* s=)
-				fil_system->read_page0(space->id);
+#ifndef DBUG_OFF
+			fil_space_t* s= fil_system.read_page0(
+				space->id);
 			ut_ad(!s || s == space);
+#endif
 			space->n_pending_ops--;
 			if (!space->size) {
 				/* Page 0 was not loaded.
@@ -2303,8 +2330,7 @@ static void fil_crypt_rotation_list_fill()
 			}
 		}
 
-		fil_system->rotation_list.push_back(*space);
-		space->is_in_rotation_list = true;
+		UT_LIST_ADD_LAST(fil_system.rotation_list, space);
 	}
 }
 
@@ -2316,12 +2342,12 @@ void
 fil_crypt_set_rotate_key_age(
 	uint	val)
 {
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 	srv_fil_crypt_rotate_key_age = val;
 	if (val == 0) {
 		fil_crypt_rotation_list_fill();
 	}
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 	os_event_set(fil_crypt_threads_event);
 }
 
@@ -2345,7 +2371,7 @@ void
 fil_crypt_set_encrypt_tables(
 	uint val)
 {
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 
 	srv_encrypt_tables = val;
 
@@ -2353,7 +2379,7 @@ fil_crypt_set_encrypt_tables(
 		fil_crypt_rotation_list_fill();
 	}
 
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 
 	os_event_set(fil_crypt_threads_event);
 }
@@ -2463,7 +2489,7 @@ fil_space_crypt_get_status(
 {
 	memset(status, 0, sizeof(*status));
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	/* If there is no crypt data and we have not yet read
 	page 0 for this tablespace, we need to read it before
@@ -2528,7 +2554,7 @@ fil_space_get_scrub_status(
 {
 	memset(status, 0, sizeof(*status));
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 	fil_space_crypt_t* crypt_data = space->crypt_data;
 
 	status->space = space->id;
