@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2019, MariaDB Corporation.
+Copyright (c) 2018, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -38,6 +38,7 @@ Created 2013/03/27 Allen Lai and Jimmy Yang
 #include "trx0undo.h"
 #include "srv0mon.h"
 #include "gis0geo.h"
+#include <cmath>
 
 /*************************************************************//**
 Initial split nodes info for R-tree split.
@@ -71,7 +72,7 @@ rtr_page_split_initialize_nodes(
 	page = buf_block_get_frame(block);
 	n_uniq = dict_index_get_n_unique_in_tree(cursor->index);
 
-	n_recs = page_get_n_recs(page) + 1;
+	n_recs = ulint(page_get_n_recs(page)) + 1;
 
 	/*We reserve 2 MBRs memory space for temp result of split
 	algrithm. And plus the new mbr that need to insert, we
@@ -86,7 +87,7 @@ rtr_page_split_initialize_nodes(
 	stop = task + n_recs;
 
 	rec = page_rec_get_next(page_get_infimum_rec(page));
-	ut_d(const bool is_leaf = page_is_leaf(page));
+	const bool is_leaf = page_is_leaf(page);
 	*offsets = rec_get_offsets(rec, cursor->index, *offsets, is_leaf,
 				   n_uniq, &heap);
 
@@ -133,10 +134,8 @@ rtr_index_build_node_ptr(
 					pointer */
 	ulint			page_no,/*!< in: page number to put in node
 					pointer */
-	mem_heap_t*		heap,	/*!< in: memory heap where pointer
+	mem_heap_t*		heap)	/*!< in: memory heap where pointer
 					created */
-	ulint			level)	/*!< in: level of rec in tree:
-					0 means leaf level */
 {
 	dtuple_t*	tuple;
 	dfield_t*	field;
@@ -291,7 +290,6 @@ rtr_update_mbr_field(
 	ulint		up_match = 0;
 	ulint		low_match = 0;
 	ulint		child;
-	ulint		level;
 	ulint		rec_info;
 	page_zip_des_t*	page_zip;
 	bool		ins_suc = true;
@@ -310,7 +308,7 @@ rtr_update_mbr_field(
 	page_zip = buf_block_get_page_zip(block);
 
 	child = btr_node_ptr_get_child_page_no(rec, offsets);
-	level = btr_page_get_level(buf_block_get_frame(block), mtr);
+	const bool is_leaf = page_is_leaf(block->frame);
 
 	if (new_rec) {
 		child_rec = new_rec;
@@ -319,7 +317,7 @@ rtr_update_mbr_field(
 	}
 
 	dtuple_t* node_ptr = rtr_index_build_node_ptr(
-		index, mbr, child_rec, child, heap, level);
+		index, mbr, child_rec, child, heap);
 
 	/* We need to remember the child page no of cursor2, since page could be
 	reorganized or insert a new rec before it. */
@@ -429,7 +427,7 @@ rtr_update_mbr_field(
 		ut_ad(old_rec != insert_rec);
 
 		page_cur_position(old_rec, block, &page_cur);
-		offsets2 = rec_get_offsets(old_rec, index, NULL, !level,
+		offsets2 = rec_get_offsets(old_rec, index, NULL, is_leaf,
 					   ULINT_UNDEFINED, &heap);
 		page_cur_delete_rec(&page_cur, index, offsets2, mtr);
 
@@ -459,7 +457,7 @@ update_mbr:
 
 			cur2_rec = cursor2->page_cur.rec;
 			offsets2 = rec_get_offsets(cur2_rec, index, NULL,
-						   !level,
+						   is_leaf,
 						   ULINT_UNDEFINED, &heap);
 
 			cur2_rec_info = rec_get_info_bits(cur2_rec,
@@ -519,7 +517,7 @@ update_mbr:
 		if (ins_suc) {
 			btr_cur_position(index, insert_rec, block, cursor);
 			offsets = rec_get_offsets(insert_rec,
-						  index, offsets, !level,
+						  index, offsets, is_leaf,
 						  ULINT_UNDEFINED, &heap);
 		}
 
@@ -534,7 +532,7 @@ update_mbr:
 			cur2_rec = btr_cur_get_rec(cursor2);
 
 			offsets2 = rec_get_offsets(cur2_rec, index, NULL,
-						   !level,
+						   is_leaf,
 						   ULINT_UNDEFINED, &heap);
 
 			/* If the cursor2 position is on a wrong rec, we
@@ -548,7 +546,7 @@ update_mbr:
 				while (!page_rec_is_supremum(cur2_rec)) {
 					offsets2 = rec_get_offsets(cur2_rec, index,
 								   NULL,
-								   !level,
+								   is_leaf,
 								   ULINT_UNDEFINED,
 								   &heap);
 					cur2_pno = btr_node_ptr_get_child_page_no(
@@ -634,7 +632,6 @@ rtr_adjust_upper_level(
 	buf_block_t*	new_block,	/*!< in/out: the new half page */
 	rtr_mbr_t*	mbr,		/*!< in: MBR on the old page */
 	rtr_mbr_t*	new_mbr,	/*!< in: MBR on the new page */
-	ulint		direction,	/*!< in: FSP_UP or FSP_DOWN */
 	mtr_t*		mtr)		/*!< in: mtr */
 {
 	page_t*		page;
@@ -653,10 +650,8 @@ rtr_adjust_upper_level(
 	ulint		next_page_no;
 	ulint		space;
 	page_cur_t*	page_cursor;
-	rtr_mbr_t	parent_mbr;
 	lock_prdt_t	prdt;
 	lock_prdt_t	new_prdt;
-	lock_prdt_t	parent_prdt;
 	dberr_t		err;
 	big_rec_t*	dummy_big_rec;
 	rec_t*		rec;
@@ -668,9 +663,8 @@ rtr_adjust_upper_level(
 	cursor.thr = sea_cur->thr;
 
 	/* Get the level of the split pages */
-	level = btr_page_get_level(buf_block_get_frame(block), mtr);
-	ut_ad(level
-	      == btr_page_get_level(buf_block_get_frame(new_block), mtr));
+	level = btr_page_get_level(buf_block_get_frame(block));
+	ut_ad(level == btr_page_get_level(buf_block_get_frame(new_block)));
 
 	page = buf_block_get_frame(block);
 	page_no = block->page.id.page_no();
@@ -687,8 +681,6 @@ rtr_adjust_upper_level(
 
 	page_cursor = btr_cur_get_page_cur(&cursor);
 
-	rtr_get_mbr_from_rec(page_cursor->rec, offsets, &parent_mbr);
-
 	rtr_update_mbr_field(&cursor, offsets, NULL, page, mbr, NULL, mtr);
 
 	/* Already updated parent MBR, reset in our path */
@@ -704,7 +696,7 @@ rtr_adjust_upper_level(
 	node_ptr_upper = rtr_index_build_node_ptr(
 		index, new_mbr,
 		page_rec_get_next(page_get_infimum_rec(new_page)),
-		new_page_no, heap, level);
+		new_page_no, heap);
 
 	ulint	up_match = 0;
 	ulint	low_match = 0;
@@ -750,11 +742,9 @@ rtr_adjust_upper_level(
 	prdt.op = 0;
 	new_prdt.data = static_cast<void*>(new_mbr);
 	new_prdt.op = 0;
-	parent_prdt.data = static_cast<void*>(&parent_mbr);
-	parent_prdt.op = 0;
 
 	lock_prdt_update_parent(block, new_block, &prdt, &new_prdt,
-				&parent_prdt, dict_index_get_space(index),
+				index->table->space_id,
 				page_cursor->block->page.id.page_no());
 
 	mem_heap_free(heap);
@@ -921,7 +911,7 @@ rtr_split_page_move_rec_list(
 	same temp-table in parallel.
 	max_trx_id is ignored for temp tables because it not required
 	for MVCC. */
-	if (is_leaf && !dict_table_is_temporary(index->table)) {
+	if (is_leaf && !index->table->is_temporary()) {
 		page_update_max_trx_id(new_block, NULL,
 				       page_get_max_trx_id(page),
 				       mtr);
@@ -1009,7 +999,6 @@ rtr_page_split_and_insert(
 	page_t*			page;
 	page_t*			new_page;
 	ulint			page_no;
-	byte			direction;
 	ulint			hint_page_no;
 	buf_block_t*		new_block;
 	page_zip_des_t*		page_zip;
@@ -1055,7 +1044,7 @@ func_start:
 	block = btr_cur_get_block(cursor);
 	page = buf_block_get_frame(block);
 	page_zip = buf_block_get_page_zip(block);
-	page_level = btr_page_get_level(page, mtr);
+	page_level = btr_page_get_level(page);
 	current_ssn = page_get_ssn_id(page);
 
 	ut_ad(mtr_memo_contains(mtr, block, MTR_MEMO_PAGE_X_FIX));
@@ -1063,7 +1052,7 @@ func_start:
 
 	page_no = block->page.id.page_no();
 
-	if (btr_page_get_prev(page, mtr) == FIL_NULL && !page_is_leaf(page)) {
+	if (!page_has_prev(page) && !page_is_leaf(page)) {
 		first_rec = page_rec_get_next(
 			page_get_infimum_rec(buf_block_get_frame(block)));
 	}
@@ -1073,7 +1062,7 @@ func_start:
 		*heap, cursor, offsets, tuple, &buf_pos);
 
 	/* Divide all mbrs to two groups. */
-	n_recs = page_get_n_recs(page) + 1;
+	n_recs = ulint(page_get_n_recs(page)) + 1;
 
 	end_split_node = rtr_split_node_array + n_recs;
 
@@ -1099,9 +1088,8 @@ func_start:
 					   static_cast<uchar*>(first_rec));
 
 	/* Allocate a new page to the index */
-	direction = FSP_UP;
 	hint_page_no = page_no + 1;
-	new_block = btr_page_alloc(cursor->index, hint_page_no, direction,
+	new_block = btr_page_alloc(cursor->index, hint_page_no, FSP_UP,
 				   page_level, mtr, mtr);
 	new_page_zip = buf_block_get_page_zip(new_block);
 	btr_page_create(new_block, new_page_zip, cursor->index,
@@ -1276,12 +1264,12 @@ after_insert:
 
 	/* Check any predicate locks need to be moved/copied to the
 	new page */
-	lock_prdt_update_split(block, new_block, &prdt, &new_prdt,
-			       dict_index_get_space(cursor->index), page_no);
+	lock_prdt_update_split(new_block, &prdt, &new_prdt,
+			       cursor->index->table->space_id, page_no);
 
 	/* Adjust the upper level. */
 	rtr_adjust_upper_level(cursor, flags, block, new_block,
-			       &mbr, &new_mbr, direction, mtr);
+			       &mbr, &new_mbr, mtr);
 
 	/* Save the new ssn to the root page, since we need to reinit
 	the first ssn value from it after restart server. */
@@ -1305,7 +1293,7 @@ after_insert:
 	if (!rec) {
 		/* We play safe and reset the free bits for new_page */
 		if (!dict_index_is_clust(cursor->index)
-		    && !dict_table_is_temporary(cursor->index->table)) {
+		    && !cursor->index->table->is_temporary()) {
 			ibuf_reset_free_bits(new_block);
 			ibuf_reset_free_bits(block);
 		}
@@ -1342,7 +1330,6 @@ dberr_t
 rtr_ins_enlarge_mbr(
 /*================*/
 	btr_cur_t*		btr_cur,	/*!< in: btr cursor */
-	que_thr_t*		thr,		/*!< in: query thread */
 	mtr_t*			mtr)		/*!< in: mtr */
 {
 	dberr_t			err = DB_SUCCESS;
@@ -1453,7 +1440,7 @@ rtr_page_copy_rec_list_end_no_locks(
 
 	btr_assert_not_corrupted(new_block, index);
 	ut_a(page_is_comp(new_page) == page_rec_is_comp(rec));
-	ut_a(mach_read_from_2(new_page + UNIV_PAGE_SIZE - 10) == (ulint)
+	ut_a(mach_read_from_2(new_page + srv_page_size - 10) == (ulint)
 	     (page_is_comp(new_page) ? PAGE_NEW_INFIMUM : PAGE_OLD_INFIMUM));
 
 	cur_rec = page_rec_get_next(
@@ -1676,10 +1663,7 @@ rtr_merge_mbr_changed(
 	btr_cur_t*		cursor2,	/*!< in: the other cursor */
 	ulint*			offsets,	/*!< in: rec offsets */
 	ulint*			offsets2,	/*!< in: rec offsets */
-	rtr_mbr_t*		new_mbr,	/*!< out: MBR to update */
-	buf_block_t*		merge_block,	/*!< in: page to merge */
-	buf_block_t*		block,		/*!< in: page be merged */
-	dict_index_t*		index)		/*!< in: index */
+	rtr_mbr_t*		new_mbr)	/*!< out: MBR to update */
 {
 	double*		mbr;
 	double		mbr1[SPDIMS * 2];
@@ -1724,9 +1708,6 @@ rtr_merge_and_update_mbr(
 	ulint*			offsets,	/*!< in: rec offsets */
 	ulint*			offsets2,	/*!< in: rec offsets */
 	page_t*			child_page,	/*!< in: the page. */
-	buf_block_t*		merge_block,	/*!< in: page to merge */
-	buf_block_t*		block,		/*!< in: page be merged */
-	dict_index_t*		index,		/*!< in: index */
 	mtr_t*			mtr)		/*!< in: mtr */
 {
 	dberr_t			err = DB_SUCCESS;
@@ -1736,8 +1717,7 @@ rtr_merge_and_update_mbr(
 	ut_ad(dict_index_is_spatial(cursor->index));
 
 	changed = rtr_merge_mbr_changed(cursor, cursor2, offsets, offsets2,
-					&new_mbr, merge_block,
-					block, index);
+					&new_mbr);
 
 	/* Update the mbr field of the rec. And will delete the record
 	pointed by cursor2 */
@@ -1747,7 +1727,7 @@ rtr_merge_and_update_mbr(
 			err = DB_ERROR;
 		}
 	} else {
-		rtr_node_ptr_delete(cursor2->index, cursor2, block, mtr);
+		rtr_node_ptr_delete(cursor2, mtr);
 	}
 
 	return(err);
@@ -1758,10 +1738,8 @@ Deletes on the upper level the node pointer to a page. */
 void
 rtr_node_ptr_delete(
 /*================*/
-	dict_index_t*	index,	/*!< in: index tree */
 	btr_cur_t*	cursor, /*!< in: search cursor, contains information
 				about parent nodes in search */
-	buf_block_t*	block,	/*!< in: page whose node pointer is deleted */
 	mtr_t*		mtr)	/*!< in: mtr */
 {
 	ibool		compressed;
@@ -1851,12 +1829,14 @@ rtr_rec_cal_increase(
 @param[in]	tuple	range tuple containing mbr, may also be empty tuple
 @param[in]	mode	search mode
 @return estimated number of rows */
-int64_t
+ha_rows
 rtr_estimate_n_rows_in_range(
 	dict_index_t*	index,
 	const dtuple_t*	tuple,
 	page_cur_mode_t	mode)
 {
+	ut_ad(dict_index_is_spatial(index));
+
 	/* Check tuple & mode */
 	if (tuple->n_fields == 0) {
 		return(HA_POS_ERROR);
@@ -1878,64 +1858,47 @@ rtr_estimate_n_rows_in_range(
 	);
 
 	/* Read mbr from tuple. */
-	const dfield_t*	dtuple_field;
-	ulint		dtuple_f_len MY_ATTRIBUTE((unused));
 	rtr_mbr_t	range_mbr;
 	double		range_area;
 
-	dtuple_field = dtuple_get_nth_field(tuple, 0);
-	dtuple_f_len = dfield_get_len(dtuple_field);
-	const byte* range_mbr_ptr = static_cast<const byte*>(
+	const dfield_t* dtuple_field = dtuple_get_nth_field(tuple, 0);
+	ut_ad(dfield_get_len(dtuple_field) >= DATA_MBR_LEN);
+	const byte* range_mbr_ptr = reinterpret_cast<const byte*>(
 		dfield_get_data(dtuple_field));
 
-	ut_ad(dtuple_f_len >= DATA_MBR_LEN);
 	rtr_read_mbr(range_mbr_ptr, &range_mbr);
 	range_area = (range_mbr.xmax - range_mbr.xmin)
 		 * (range_mbr.ymax - range_mbr.ymin);
 
 	/* Get index root page. */
-	page_size_t	page_size(dict_table_page_size(index->table));
-	page_id_t	page_id(dict_index_get_space(index),
-				dict_index_get_page(index));
 	mtr_t		mtr;
-	buf_block_t*	block;
-	page_t*		page;
-	ulint		n_recs;
 
-	mtr_start(&mtr);
-	mtr.set_named_space(dict_index_get_space(index));
-	mtr_s_lock(dict_index_get_lock(index), &mtr);
+	mtr.start();
+	index->set_modified(mtr);
+	mtr_s_lock(&index->lock, &mtr);
 
-	block = btr_block_get(page_id, page_size, RW_S_LATCH, index, &mtr);
-	page = buf_block_get_frame(block);
-	n_recs = page_header_get_field(page, PAGE_N_RECS);
+	buf_block_t* block = btr_block_get(
+		page_id_t(index->table->space_id, index->page),
+		page_size_t(index->table->space->flags),
+		RW_S_LATCH, index, &mtr);
+	const page_t* page = buf_block_get_frame(block);
+	const unsigned n_recs = page_header_get_field(page, PAGE_N_RECS);
 
 	if (n_recs == 0) {
-		mtr_commit(&mtr);
+		mtr.commit();
 		return(HA_POS_ERROR);
 	}
 
-	rec_t*		rec;
-	byte*		field;
-	ulint		len;
-	ulint*		offsets = NULL;
-	mem_heap_t*	heap;
-
-	heap = mem_heap_create(512);
-	rec = page_rec_get_next(page_get_infimum_rec(page));
-	offsets = rec_get_offsets(rec, index, offsets, page_rec_is_leaf(rec),
-				  ULINT_UNDEFINED, &heap);
-
 	/* Scan records in root page and calculate area. */
 	double	area = 0;
-	while (!page_rec_is_supremum(rec)) {
+	for (const rec_t* rec = page_rec_get_next(
+		     page_get_infimum_rec(block->frame));
+	     !page_rec_is_supremum(rec);
+	     rec = page_rec_get_next_const(rec)) {
 		rtr_mbr_t	mbr;
 		double		rec_area;
 
-		field = rec_get_nth_field(rec, offsets, 0, &len);
-		ut_ad(len == DATA_MBR_LEN);
-
-		rtr_read_mbr(field, &mbr);
+		rtr_read_mbr(rec, &mbr);
 
 		rec_area = (mbr.xmax - mbr.xmin) * (mbr.ymax - mbr.ymin);
 
@@ -1952,8 +1915,8 @@ rtr_estimate_n_rows_in_range(
 			case PAGE_CUR_WITHIN:
 			case PAGE_CUR_MBR_EQUAL:
 				if (rtree_key_cmp(
-					PAGE_CUR_WITHIN, range_mbr_ptr,
-					DATA_MBR_LEN, field, DATA_MBR_LEN)
+					    PAGE_CUR_WITHIN, range_mbr_ptr,
+					    DATA_MBR_LEN, rec, DATA_MBR_LEN)
 				    == 0) {
 					area += 1;
 				}
@@ -1967,22 +1930,23 @@ rtr_estimate_n_rows_in_range(
 			switch (mode) {
 			case PAGE_CUR_CONTAIN:
 			case PAGE_CUR_INTERSECT:
-				area += rtree_area_overlapping(range_mbr_ptr,
-						field, DATA_MBR_LEN) / rec_area;
+				area += rtree_area_overlapping(
+					range_mbr_ptr, rec, DATA_MBR_LEN)
+					/ rec_area;
 				break;
 
 			case PAGE_CUR_DISJOINT:
 				area += 1;
-				area -= rtree_area_overlapping(range_mbr_ptr,
-						field, DATA_MBR_LEN) / rec_area;
+				area -= rtree_area_overlapping(
+					range_mbr_ptr, rec, DATA_MBR_LEN)
+					/ rec_area;
 				break;
 
 			case PAGE_CUR_WITHIN:
 			case PAGE_CUR_MBR_EQUAL:
-				if (rtree_key_cmp(
-					PAGE_CUR_WITHIN, range_mbr_ptr,
-					DATA_MBR_LEN, field, DATA_MBR_LEN)
-				    == 0) {
+				if (!rtree_key_cmp(
+					    PAGE_CUR_WITHIN, range_mbr_ptr,
+					    DATA_MBR_LEN, rec, DATA_MBR_LEN)) {
 					area += range_area / rec_area;
 				}
 
@@ -1991,17 +1955,14 @@ rtr_estimate_n_rows_in_range(
 				ut_error;
 			}
 		}
-
-		rec = page_rec_get_next(rec);
 	}
 
-	mtr_commit(&mtr);
-	mem_heap_free(heap);
+	mtr.commit();
 
-	if (!isfinite(area)) {
+	if (!std::isfinite(area)) {
 		return(HA_POS_ERROR);
 	}
 
-	return(static_cast<int64_t>(dict_table_get_n_rows(index->table)
-				    * area / n_recs));
+	area /= n_recs;
+	return ha_rows(dict_table_get_n_rows(index->table) * area);
 }
